# -*- coding: utf-8 -*-
"""
Loads configuration.
"""

from __future__ import absolute_import, unicode_literals

import io
import os

import yaml

DEFAULT_CONF = os.path.join(os.path.abspath(os.path.dirname(__file__)), "polarion_tools.yaml")
PROJECT_CONFS = ("polarion_tools.yaml", "conf/polarion_tools.yaml")
USER_CONF = os.path.expanduser("~/.config/dump2polarion.yaml")


def find_vcs_root(path, dirs=(".git",)):
    """Searches up from a given path to find the project root."""
    prev, path = None, os.path.abspath(path)
    while prev != path:
        if any(os.path.exists(os.path.join(path, d)) for d in dirs):
            return path
        prev, path = path, os.path.abspath(os.path.join(path, os.pardir))
    return None


def _get_project_conf():
    """Loads configuration from project config file."""
    default_conf = {}

    project_root = find_vcs_root(".")
    if project_root is None:
        return default_conf

    for pconf in PROJECT_CONFS:
        try:
            with io.open(os.path.join(project_root, pconf), encoding="utf-8") as input_file:
                return yaml.load(input_file)
        except EnvironmentError:
            pass

    return default_conf


def _get_user_conf():
    """Loads configuration from user config file."""
    try:
        with io.open(USER_CONF, encoding="utf-8") as input_file:
            return yaml.load(input_file)
    except EnvironmentError:
        pass

    return {}


def get_config(default_only=False):
    """Loads configuration.

    Args:
        default_only: bool if only default config will be loaded (for easier testing)
    """
    with io.open(DEFAULT_CONF, encoding="utf-8") as input_file:
        config_settings = yaml.load(input_file)

    if default_only:
        return config_settings

    # merge with project configuration
    project_conf = _get_project_conf()
    config_settings.update(project_conf)

    # merge with user configuration
    user_conf = _get_user_conf()
    config_settings.update(user_conf)

    return config_settings



# -*- coding: utf-8 -*-
"""
Parses Polarion docstring.
"""

from __future__ import absolute_import, unicode_literals

import ast
from collections import namedtuple

FORMATED_KEYS = ("setup", "teardown")

DocstringsRecord = namedtuple("DocstringsRecord", "lineno column value")


# pylint: disable=too-few-public-methods
class SECTIONS(object):
    """Valid sections in Polarion docstring."""

    polarion = "Polarion"
    steps = "testSteps"
    results = "expectedResults"


def _get_section_start(doc_list, section):
    """Finds the line with "section" (e.g. "Polarion", "testSteps", etc.)."""
    section = "{}:".format(section)
    for index, line in enumerate(doc_list):
        if section != line.strip():
            continue
        indent = len(line) - len(line.lstrip(" "))
        return index + 1, indent
    return None, None


def _get_key_value(line):
    """Gets the key and value out of docstring line."""
    data = line.split(":")
    if len(data) == 1:
        data.append("")

    key = data[0].strip()

    value = ":".join(data[1:]).strip()
    if value == "None":
        value = None

    return key, value


# pylint: disable=too-many-locals
def _lines_to_dict(lines, start=0, lineno_offset=0, stop=None):
    """Creates dictionary out of docstring lines.

    Includes column and line number info for each record.
    """
    if stop:
        lines = lines[start:stop]
    else:
        lines = lines[start:]

    lines_dict = {}
    indent = len(lines[0]) - len(lines[0].lstrip(" "))
    prev_key = None
    for num, line in enumerate(lines, 1):
        line_stripped = line.strip()
        if not line_stripped:
            break

        curr_indent = len(line) - len(line.lstrip(" "))

        if curr_indent < indent:
            break

        # check if the line should be appended to previous key
        first_word = line_stripped.split(" ")[0] or line_stripped
        if prev_key and curr_indent > indent and first_word[-1] != ":":
            sep = "\n" if prev_key in FORMATED_KEYS else " "
            prev_lineno, prev_indent, prev_value = lines_dict[prev_key]
            lines_dict[prev_key] = DocstringsRecord(
                prev_lineno, prev_indent, "{}{}{}".format(prev_value, sep, line_stripped)
            )
            continue
        else:
            prev_key = None

        if curr_indent > indent:
            continue

        key, value = _get_key_value(line)
        lines_dict[key] = DocstringsRecord(num + lineno_offset, indent, value)
        prev_key = key

    return lines_dict


def _lines_to_list(lines, start=0, lineno_offset=0, stop=None):
    """Creates list out of docstring lines.

    Includes column and line number info for each line.
    """
    if stop:
        lines = lines[start:stop]
    else:
        lines = lines[start:]

    lines_list = []
    indent = len(lines[0]) - len(lines[0].lstrip(" "))
    for num, line in enumerate(lines, 1):
        line_stripped = line.strip()
        if not line_stripped:
            continue
        curr_indent = len(line) - len(line.lstrip(" "))
        if curr_indent < indent:
            break

        # check if the line should be appended to previous key
        first_word = line_stripped.split(" ")[0] or line_stripped
        if curr_indent > indent and first_word[-1] != ":":
            prev_lineno, prev_indent, prev_value = lines_list.pop()
            lines_list.append(
                DocstringsRecord(
                    prev_lineno, prev_indent, "{} {}".format(prev_value, line_stripped)
                )
            )
            continue

        if curr_indent > indent:
            continue
        lines_list.append(DocstringsRecord(num + lineno_offset, indent, line_stripped))

    return lines_list


def parse_docstring(docstring):
    """Parses docstrings to dictionary. E.g.:

    Polarion:
        assignee: mkourim
        casecomponent: nonexistent
        testSteps:
            1. Step with really long description
               that doesn't fit into one line
            2. Do that
        expectedResults:
            1. Success outcome with really long description
               that doesn't fit into one line
            2. 2
        caseimportance: low
        title: Some test with really long description
               that doesn't fit into one line
        setup: Do this:
               - first thing
               - second thing
        foo: this is an unknown field

    This is not included.
    """
    doc_list = docstring.split("\n")

    polarion_start, __ = _get_section_start(doc_list, SECTIONS.polarion)
    if not polarion_start:
        return None

    docstring_dict = _lines_to_dict(doc_list, start=polarion_start)
    if SECTIONS.steps in docstring_dict and docstring_dict[SECTIONS.steps][2]:
        steps_start, __ = _get_section_start(doc_list, SECTIONS.steps)
        steps_list = _lines_to_list(
            doc_list, start=steps_start, lineno_offset=steps_start - polarion_start
        )
        docstring_dict[SECTIONS.steps] = steps_list
    if SECTIONS.results in docstring_dict and docstring_dict[SECTIONS.results][2]:
        results_start, __ = _get_section_start(doc_list, SECTIONS.results)
        results_list = _lines_to_list(
            doc_list, start=results_start, lineno_offset=results_start - polarion_start
        )
        docstring_dict[SECTIONS.results] = results_list

    return docstring_dict


def _get_tree(filename):
    with open(filename) as infile:
        source = infile.read()

    tree = ast.parse(source, filename=filename)
    return tree


def strip_polarion_data(docstring):
    """Removes the Polarion section out of the docstring."""
    docstring_list = docstring.split("\n")
    new_docstring_list = []
    indent = 0
    polarion_section = "{}:".format(SECTIONS.polarion)
    for line in docstring_list:
        if line.strip() == polarion_section:
            indent = len(line) - len(line.lstrip(" "))
            continue
        if indent:
            curr_indent = len(line) - len(line.lstrip(" "))
            if curr_indent > indent:
                continue
        new_docstring_list.append(line)
    new_docstring = "\n".join(new_docstring_list).replace("\n\n", "\n")
    return new_docstring


def get_docstring_from_func(func_node):
    """Gets docstring from function definition."""
    docstring = None
    try:
        if (
            func_node.body
            and isinstance(func_node.body[0], ast.Expr)
            and isinstance(func_node.body[0].value, ast.Str)
        ):
            docstring = func_node.body[0].value.s
    # pylint: disable=broad-except
    except Exception:
        return None

    # for Python2
    try:
        docstring = docstring.decode("utf-8")
    except AttributeError:
        pass

    return docstring


def get_docstrings_in_file(tree, filename, tests_only=True):
    """Returns parsed Polarion docstrings present in the python source file."""
    if not tree:
        tree = _get_tree(filename)

    docstrings = []
    for node in ast.walk(tree):
        if not isinstance(node, ast.FunctionDef):
            continue

        if tests_only and node.name.find("test_") != 0:
            continue

        docstring = get_docstring_from_func(node)

        # test doesn't have docstring, i.e. it's missing also the Polarion section
        if not docstring:
            docstrings.append(DocstringsRecord(node.body[0].lineno - 1, node.col_offset, {}))
            continue

        doc_list = docstring.split("\n")
        docstring_start = node.body[0].lineno - len(doc_list)
        polarion_start, polarion_column = _get_section_start(doc_list, SECTIONS.polarion)

        if not polarion_start:
            # docstring is missing the Polarion section
            docstrings.append(DocstringsRecord(docstring_start + 1, node.col_offset + 4, {}))
            continue

        polarion_offset = docstring_start + polarion_start
        docstrings.append(
            DocstringsRecord(polarion_offset, polarion_column, parse_docstring(docstring))
        )

    return docstrings



# -*- coding: utf-8 -*-
"""
Checks Polarion docstrings.
"""

from __future__ import absolute_import, print_function, unicode_literals

import os
import sys
from collections import namedtuple

from pkg_resources import DistributionNotFound, get_distribution

from polarion_docstrings import configuration, parser

DocstringsError = namedtuple("DocstringsError", "lineno column message func")
FieldRecord = namedtuple("FieldRecord", "lineno column field")


def _validate(docstring_dict, key, valid_values):
    record = docstring_dict.get(key)
    if record is not None:
        return record.value in valid_values[key]
    return True


def _get_unknown_fields(docstring_dict, known_fields):
    unknown = [
        FieldRecord(docstring_dict[key].lineno, docstring_dict[key].column, key)
        for key in docstring_dict
        if key not in known_fields
    ]
    return unknown


def _get_invalid_fields(docstring_dict, valid_values):
    results = {key: _validate(docstring_dict, key, valid_values) for key in valid_values}
    invalid = [
        FieldRecord(docstring_dict[key].lineno, docstring_dict[key].column, key)
        for key, result in results.items()
        if not result
    ]
    return invalid


def _get_missing_fields(docstring_dict, required_keys):
    if not required_keys:
        return []
    missing = [key for key in required_keys if key not in docstring_dict]
    return missing


def _get_markers_fields(docstring_dict, marker_fields):
    if not marker_fields:
        return []
    markers = [
        FieldRecord(docstring_dict[key].lineno, docstring_dict[key].column, key)
        for key in marker_fields
        if key in docstring_dict
    ]
    return markers


def validate_docstring(docstring_dict, config):
    """Returns tuple with lists of problematic fields."""
    cfg_docstrings = config["docstrings"]
    unknown = _get_unknown_fields(docstring_dict, cfg_docstrings.get("known_fields"))
    invalid = _get_invalid_fields(docstring_dict, cfg_docstrings.get("valid_values"))
    missing = _get_missing_fields(docstring_dict, cfg_docstrings.get("required_fields"))
    markers = _get_markers_fields(docstring_dict, cfg_docstrings.get("marker_fields"))
    return unknown, invalid, missing, markers


def get_fields_errors(validated_docstring, docstring_dict, marker_fields, lineno=0, column=0):
    """Produces fields errors for the flake8 checker."""
    errors = []
    func = polarion_checks492
    unknown, invalid, missing, markers = validated_docstring

    for num, col, field in unknown:
        errors.append(
            DocstringsError(lineno + num, col, 'P666 Unknown field "{}"'.format(field), func)
        )
    for num, col, field in invalid:
        errors.append(
            DocstringsError(
                lineno + num,
                col,
                'P667 Invalid value "{}" of the "{}" field'.format(
                    docstring_dict[field].value, field
                ),
                func,
            )
        )
    for num, col, field in markers:
        errors.append(
            DocstringsError(
                lineno + num,
                col,
                'P668 Field "{}" should be handled by the "{}" marker'.format(
                    field, marker_fields.get(field)
                ),
                func,
            )
        )
    for field in missing:
        errors.append(
            DocstringsError(lineno, column, 'P669 Missing required field "{}"'.format(field), func)
        )

    if errors:
        errors = sorted(errors, key=lambda tup: tup[0])
    return errors


def print_errors(errors):
    """Prints errors without using flake8."""
    for err in errors:
        print("line: {}:{}: {}".format(err.lineno, err.column, err.message), file=sys.stderr)


def check_docstrings(docstrings, config):
    """Runs checks on each docstring."""
    errors = []
    for record in docstrings:
        if record.value:
            valdoc = validate_docstring(record.value, config)
            errors.extend(
                get_fields_errors(
                    valdoc,
                    record.value,
                    config["docstrings"].get("marker_fields") or {},
                    record.lineno,
                    record.column,
                )
            )
        else:
            errors.append(
                DocstringsError(
                    record.lineno,
                    record.column,
                    'P665 Missing "Polarion" section',
                    polarion_checks492,
                )
            )
    return errors


def run_checks(tree, filename, config):
    """Checks docstrings in python source file."""
    docstrings = parser.get_docstrings_in_file(tree, filename)
    errors = check_docstrings(docstrings, config)
    return errors


def polarion_checks492(tree, filename):
    """The flake8 entry point."""
    abs_filename = os.path.abspath(filename)
    __, tail = os.path.split(abs_filename)

    config = configuration.get_config()
    cfg_valid = config.get("docstrings") or {}
    cfg_valid = cfg_valid.get("valid_values")
    if not cfg_valid:
        return []  # must be iterable

    # check only test files under tests path
    if not (
        config.get("tests_relative_path", "tests") in abs_filename and tail.startswith("test_")
    ):
        return []

    return run_checks(tree, filename, config)


try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    __version__ = "0.0"

polarion_checks492.name = "polarion_checks"
polarion_checks492.version = __version__



# encoding: utf-8
# pylint: disable=missing-docstring

from __future__ import unicode_literals

from polarion_docstrings import checker, configuration

RESULTS = [
    (50, 0, 'P665 Missing "Polarion" section'),
    (54, 0, 'P665 Missing "Polarion" section'),
    (60, 4, 'P665 Missing "Polarion" section'),
    (10, 4, 'P665 Missing "Polarion" section'),
    (14, 8, 'P665 Missing "Polarion" section'),
    (21, 8, 'P669 Missing required field "initialEstimate"'),
    (23, 12, 'P667 Invalid value "nonexistent" of the "casecomponent" field'),
    (39, 12, 'P667 Invalid value "level1" of the "caselevel" field'),
    (39, 12, 'P668 Field "caselevel" should be handled by the "@pytest.mark.tier" marker'),
    (40, 12, 'P668 Field "caseautomation" should be handled by the "@pytest.mark.manual" marker'),
    (
        41,
        12,
        'P668 Field "linkedWorkItems" should be handled by the "@pytest.mark.requirements" marker',
    ),
    (42, 12, 'P666 Unknown field "foo"'),
]


def _strip_func(errors):
    return [(lineno, col, msg) for lineno, col, msg, __ in errors]


def test_checker(source_file):
    config = configuration.get_config(default_only=True)
    errors = checker.run_checks(None, source_file, config)
    errors = _strip_func(errors)
    assert len(errors) == len(RESULTS)
    assert errors == RESULTS
